var _ = require('lodash');
const webpack = require("webpack");
const merge = require('webpack-merge');
var rimraf = require('rimraf');
var AWS = require('aws-sdk');
var s3 = new AWS.S3();
const { readyBundle } = require('./helpers/ready-bundle');
const webpackProd = require('./webpack.prod');
const webpackDev = require('./webpack.dev');
const webpackServer = require('./webpack.server');

// builds app
function build(app) {
  console.log('Building Zealder bundles...');

  let promises = [];

  // make server bundles
  if ('production' === app.get('env')) {
    bundleServer(app, promises);
  }

  // make bundle for all sites
  bundleSites(app, promises);

  Promise.all(promises).then(function() {
    app.debug('build completed!');
    // all done
    return process.exit();
  });
}
module.exports = build;

// make server bundles
function bundleServer(app, promises) {
  var { themes, config } = app.locals;

  // iterate all themes
  _.forEach(themes, function(val, theme) {
    promises.push( new Promise(function(resolve) {
      // make bundle config for server-renderer
      readyBundle(app, '', theme, resolve, (wpConfig, sitePath, appBundled, done) => {
        // if wpConfig is empty that means packages was not changed, so skip
        if (!wpConfig) return;

        // Webpack specific for bundled file
        let conf = {
          output: {
            filename: `${config.contentPath}/build/${theme}/server-render.js`
          },
          entry: `${themes[appBundled].path}/${themes[appBundled].zealderConfig.serverRenderFile}`
        };

        var webpackConfig = merge(webpackServer, wpConfig, conf);
        webpack(webpackConfig, (err, stats) => {
          if (err) throw new Error(err.stack || err);
          const info = stats.toJson();
          if (stats.hasErrors()) app.error(info.errors);
          done();
        });
      });
    }));
  });
}

function bundleSites(app, promises) {
  var { config } = app.locals;

  // configure data-source
  var DataSource = require('loopback-datasource-juggler').DataSource;
  var mainDs = new DataSource(config.db);

  promises.push( new Promise(function(resolve) {
    // get all sites to bundle
    let sql = "SELECT * FROM zlSite WHERE zsTheme != ?";
    mainDs.connector.execute(sql, [''], function (err, zs) {
      let sitePromises = [];
      if (err) app.error(err);
      else {
        _.forEach(zs, function(zlSite) {
          sitePromises.push( new Promise(function(siteResolve) {
            let siteRoot = (zlSite.zsBaseUrlPath)?`/${zlSite.zsBaseUrlPath}/`:`/`;

            // make bundle config for site
            readyBundle(app, zlSite, '', siteResolve, (wpConfig, sitePath, appBundled, done) => {
              // if wpConfig is empty that means packages was not changed, so skip
              if (!wpConfig) return;

              // rmove existing /dist then build
              rimraf(`${sitePath}/dist`, (err) => {
                if (err) app.error(err);

                // Webpack specific for bundled file
                let conf = {
                  output: {
                    path: `${sitePath}/dist`,
                    publicPath: siteRoot
                  }
                };

                var envWp = webpackDev;
                if ('production' === app.get('env')) envWp = webpackProd;

                var webpackConfig = merge(envWp, wpConfig, conf);
                webpack(webpackConfig, (err, stats) => {
                  if (err) throw new Error(err.stack || err);
                  const info = stats.toJson();
                  if (stats.hasErrors()) app.error(info.errors);
                  done();
                });
              });
            });
          }));
        });
      }
      Promise.all(sitePromises).then(function() {
        if ('production' === app.get('env')) {
          // set delete date for expired files
          var date = new Date();
          // set to midnight gmt
          date.setHours(0,0,0,0);
          // TODO: this may only work where timezone is negative gmt
          var d = new Date(date.valueOf() - date.getTimezoneOffset() * 60000);
          d.setDate(d.getDate() + 2);
          app.debug('Updating LifecycleConfiguration to delete on: %s', d.toString());
          var params = {
            Bucket: process.env.ZEAL_S3_BUCKET,
            LifecycleConfiguration: {
              Rules: [
                {
                  Expiration: {
                    Date: d
                  },
                  Filter: {
                    Tag: {
                      Key: "pending-delete",
                      Value: "true"
                    }
                  },
                  ID: "ZealderDelete",
                  Status: "Enabled"
                }
              ]
            }
          };
          s3.putBucketLifecycleConfiguration(params, function(err, data) {
            if (err) {
              app.error(`unable to update S3 BucketLifecycleConfiguration %O`, err);
            }
            resolve();
          });
        }
        else {
          resolve();
        }
      });
    });
  }));
}
