// @flow
'use strict';

var ExtractTextPlugin = require('extract-text-webpack-plugin');
var path = require('path');
var fs = require('fs');
var mkdirp = require('mkdirp');
var _ = require('lodash');
var exec = require('child_process').exec;
var AWS = require('aws-sdk');
var s3 = new AWS.S3();
var _s3 = require('s3');
var s3Client = _s3.createClient();

// for getting thems/apps ready for bundling
function readyBundleConf(app, wpConfig, modIncl, compDict, sassPath, appBundled, wpEntry, type, key, pack, packagesVer) {
  app.debug('preparing bundle for  %s', key);

  // lets theme use alias like "import widget from 'theme-name/widget'"
  wpConfig.resolve.alias[key] = `${pack.path}`;

  packagesVer[key] = pack.version;

  // limit path for loaders
  modIncl.push(new RegExp(`${key}/(?!.*node_modules)`));

  if (pack.zealderConfig) {
    // make dictionary of custom react Components
    if (pack.zealderConfig.reactComponent) {
      compDict += `"${key}": require('${pack.path}/${pack.zealderConfig.reactComponent}'),`;
    }

    // site bundle entries
    if (pack.zealderConfig.vendorJsFile) {
      if (!wpEntry.vendor) wpEntry.vendor = [];
      wpEntry.vendor.push(`${pack.path}/${pack.zealderConfig.vendorJsFile}`);
    }

    if (pack.zealderConfig.cssStyle) {
      // add style in opposite order so higher theme will have more authority
      wpEntry.app.unshift(`${pack.path}/${pack.zealderConfig.cssStyle}`);
    }

    if (pack.zealderConfig.sassPath) {
      /* when a theme needs to push sassPath up the chain. Maybe to
      make available to a framework/lib such as bootstrap */
      sassPath.push(`${pack.path}/${pack.zealderConfig.sassPath}`);
    }

    if (!appBundled && pack.zealderConfig.clientRenderFile
      && pack.zealderConfig.serverRenderFile && 'theme' == type ) {
      // first clientRenderFile found will be used; most themes shouldn't
      // need this if they use React/Redux becasue default theme is
      // sufficient. Themes that don't use/require React/Redux should
      // have own clientRenderFile and serverRenderFile
      wpEntry.app.push(`${pack.path}/${pack.zealderConfig.clientRenderFile}`);

      appBundled = key;
    }
  }
  // return these values
  return { compDict, appBundled };
}
exports.readyBundleConf = readyBundleConf;

function wrapup(sassPath, wpConfig, modIncl) {
  let sassOpt = { sourceMap: true }
  // if we have sass path that must be included globally, do so
  if (sassPath.length > 0) {
    sassOpt.includePaths = sassPath;
  }

  // add babel-loader
  wpConfig.module.rules.push(
    {
      test: /\.js$/,
      include: modIncl,
      loader: "babel-loader"
    },
    {
      test: /\.scss$/,
      loader: ExtractTextPlugin.extract([
        "css-loader?sourceMap",
        { loader: 'postcss-loader', options: {
          sourceMap: true,
          plugins: [
            require('autoprefixer')
          ]
         }
       },
        { loader: "sass-loader", options: sassOpt }
      ])
    }
  );
}
exports.wrapup = wrapup;

function publish(app, zlSite, contentPath, versFile, packagesVer, resolve) {
  if (zlSite && process.env.ZEAL_S3_BUCKET && 'production' === app.get('env')) {
    // move dist files to S3
    var zsBaseUrlPath = '';
    if (zlSite.zsBaseUrlPath) {
      zsBaseUrlPath = `${zlSite.zsBaseUrlPath}/`;
    }
    let distrDirs = ['css', 'js', 'fonts'];
    let promises = [];
    _.forEach(distrDirs, function(dir) {
      let localDir = `${contentPath}/dist/${dir}`;
      promises.push( new Promise(function(s3Resolve) {
        // get list of existing objects
        var listParams = {
          Bucket: process.env.ZEAL_S3_BUCKET,
          Prefix: `site/${zlSite.zsId}/${dir}/`
        };
        s3.listObjectsV2(listParams, function(err, data) {
          if (err) {
            app.error("unable to get existing S3 from: %s %O", listParams.Prefix, err);
            throw new Error('S3 fetch failed');
          }
          else {
            // make key of existing files
            let s3Exist = [];
            if (data && data.Contents) {
              data.Contents.forEach(s3File => {
                s3Exist.push(s3File.Key);
              });
            }
            // get all files that need to upload
            fs.readdir(localDir, function (err, files) {
              if (err) {
                app.error(`unable to get local files in ${localDir}`);
                throw new Error('local fetch failed');
              }
              else if (files) {
                let upPromises = [], filesKeep = [];
                files.forEach(file => {
                  let s3Key = `site/${zlSite.zsId}/${zsBaseUrlPath}${dir}/${file}`;
                  if (-1 === s3Exist.indexOf(s3Key)) {
                    upPromises.push( new Promise(function(upResolve) {
                      // now move file to S3
                      var params = {
                        localFile: `${localDir}/${file}`,
                        s3Params: {
                          Bucket: process.env.ZEAL_S3_BUCKET,
                          Key: s3Key,
                          ACL: 'public-read',
                          CacheControl: 'public, max-age=31536000, immutable',
                          StorageClass: 'ONEZONE_IA'
                        },
                      };

                      var uploader = s3Client.uploadFile(params);
                      uploader.on('error', function(err) {
                        app.error("unable to upload %s to S3. Error: %O", err);
                        throw new Error('S3 upload failed');
                      });
                      uploader.on('end', function() {
                        app.debug('uploaded: %s', s3Key);
                        upResolve();
                      });
                    }));
                  }
                  else filesKeep.push(s3Key);
                });

                // add deletion tag for files that is no longer needed
                if (s3Exist) {
                  s3Exist.forEach(s3File => {
                    if (-1 === filesKeep.indexOf(s3File)) {
                      upPromises.push( new Promise(function(upResolve) {
                        // tag it
                        var params = {
                          Bucket: process.env.ZEAL_S3_BUCKET,
                          Key: s3File,
                          Tagging: {
                            TagSet: [
                              {
                                Key: "pending-delete",
                                Value: "true"
                              }
                            ]
                          }
                        };
                        s3.putObjectTagging(params, function(err, data) {
                          if (err) {
                            app.error("unable to tag S3 file: %s %O", s3File, err);
                          }
                          else {
                            app.debug('S3 file tagged for delete: %s', s3File);
                          }
                          upResolve();
                        });
                      }));
                    }
                  });
                }

                Promise.all(upPromises).then(function() {
                  // delete local then resolve
                  exec('rm -r ' + localDir, s3Resolve);
                });
              }
              else s3Resolve();
            });
          }
        });
      }));
    });

    Promise.all(promises).then(function() {
      // save list of package versions in this bundle (after cb owner saves it)
      fs.writeFile(versFile, JSON.stringify(packagesVer), resolve);
    });
  }
  else fs.writeFile(versFile, JSON.stringify(packagesVer), resolve);

}

// make bundle for a site
function readyBundle(app, zlSite, theme, resolve = ()=>{}, cb) {
  var { apps, themes, cache, getInheritTheme, defaultTheme, config } = app.locals;

  var contentPath, bundlTheme;

  if (zlSite) {
    // when bundling a site (front-end)
    contentPath = `${path.join(process.cwd(), config.contentPath)}/site/${zlSite.zsId}`;
    bundlTheme = zlSite.zsTheme;
    app.debug('getting site:%s ready to bundle', zlSite.zsId);
  }
  else if (theme) {
    // when bundling a theme (server-side)
    contentPath = `${path.join(process.cwd(), config.contentPath)}/build/${theme}`;
    bundlTheme = theme;
    app.debug('getting theme:%s ready to bundle', theme);
  }

  var appBundled = 0, sassPath = [], wpEntry = { app: []};
  var wpConfig = {
    resolve: {
      alias: {
        zlSite: contentPath,
        'zealder-cms': app.locals.zlConfig.path
      }
    },
    module: { rules: [] }
  };

  // prevent from bundling 2 copies of these libraries by aliasing copy in cms
  _.forEach(app.locals.dependencies, function(ver, lib) {
    if ('production' === app.get('env')) {
      wpConfig.resolve.alias[lib] = path.join(process.cwd(), `node_modules/${lib}`)
    }
    else {
      wpConfig.resolve.alias[lib] = `${app.locals.zlConfig.path}/node_modules/${lib}`;
    }
  });

  // dictionary of components
  let compDict = 'module.exports = function() { return{';
  //
  let modIncl = [new RegExp(`zealder-cms/(?!.*node_modules)`)];

  // keep list of packages/version starting with zealder-cms
  let packagesVer = {"zealder-cms": app.locals.zlConfig.version};

  function bundleIt(type, key, pack) {
    var bdlData = readyBundleConf(app, wpConfig, modIncl, compDict, sassPath, appBundled, wpEntry, type, key, pack, packagesVer);
    if (bdlData) {
      compDict = bdlData.compDict;
      appBundled = bdlData.appBundled;
    }
  }

  // ready all themes in tree
  if (bundlTheme) {
    var themeTree = getInheritTheme(bundlTheme);
    _.forEach(themeTree, function(v, key) {
      if (appBundled || key === defaultTheme) {
        // index rendered must be last in bundle chain
        // and we dont bundle defaultTheme yet
        return false;
      }
      bundleIt('theme', key, themes[key]);
    });
  }

  // ready all site's apps
  if (!zlSite) {
    _.forEach(apps, function(v, key) {
      bundleIt('app', key, apps[key]);
    });
  }
  else if (zlSite.zsApps) {
    // we convert to json manually since we fetched directly from db
    let zsApps = JSON.parse(zlSite.zsApps);
    _.forEach(zsApps, function(key) {
      if (apps[key]) {
        bundleIt('app', key, apps[key]);
      }
    });
  }

  // lastly, bundle default theme
  if (!appBundled) {
    bundleIt('theme', defaultTheme, themes[defaultTheme]);
  }

  // add entry
  if (zlSite) wpConfig.entry = wpEntry;

  // save disctionary of components
  compDict += '}}';
  // now wrapup config
  wrapup(sassPath, wpConfig, modIncl);

  var dictFile = contentPath+'/compDict.js';
  var versFile = contentPath+'/packagesVer.js';

  // check if existing bundle was made
  if ('build-hot' !== process.env.WORK && fs.existsSync(dictFile) && fs.existsSync(versFile)) {
    let prevPackagesVer = fs.readFileSync(versFile);
    if (_.isEqual(JSON.parse(prevPackagesVer), packagesVer)) {
      // since no package in bundle has changed, keep previous one
      app.debug('skipping %s because existing packages bundle have not changed', contentPath);
      resolve();
      return cb();
    }
    else {
      // removed this file since it's outdated
      fs.unlinkSync(versFile);
    }
  }

  mkdirp.sync(contentPath);
  fs.writeFileSync(dictFile, compDict);

  // callback with wpConfig as param
  cb(wpConfig, contentPath, appBundled, () => {
    // publish it
    if ('build' === process.env.WORK) {
      publish(app, zlSite, contentPath, versFile, packagesVer, resolve);
    }
  });
}
exports.readyBundle = readyBundle;
