// @flow
'use strict';

var _ = require('lodash');

// list of roles and roles that they inherit
const ROLE_INHERIT = {
  // site contact's roles
  'Contact'          : [],
  'Contact Buyer'    : ['Contact'],
  // site staff roles
  'Site Contributor' : [],
  'Site Designer'    : ['Site Contributor'],
  'Site Agent'       : ['Site Contributor'],
  'Site CRM Rep'     : ['Site Contributor'],
  'Site Editor'      : ['Site Contributor'],
  'Site Admin'       : ['Site Editor', 'Site CRM Rep', 'Site CRM Rep', 'Site Agent', 'Site Designer'],
  'Site Owner'       : ['Site Admin'],
  // internal network staff roles
  'Network Contributor': [],
  'Network App'      : [],
  'Network Admin'    : ['Network Contributor', 'Site Owner'],
  'Network Owner'    : ['Network Admin', 'Network App']
}

function setup(app) {
  // must make sure packages are loaded first, in case a package's boot script
  // creates/updates model
  app.on('zealder packages loaded', () => {
    if ('development' === process.env.NODE_ENV || 'build' === process.env.WORK){
      // make sure all models have been created/updated
      app.dataSources.mainDs.autoupdate();
    }

    // setup for roles
    var Role = app.models.Role;

    // make these functions global to app
    app.buildAllRoles = buildAllRoles;
    app.getUserAppRoles = getUserAppRoles;

    // basic resolvers
    Role.registerResolver('Everyone', function(role, context, cb) {
      // it's a free for all
      return process.nextTick(() => cb(null, true));
    });
    Role.registerResolver('Public Only', function(role, context, cb) {
      var userId = context.accessToken.userId;
      if (userId) return process.nextTick(() => cb(null, false));
      else return process.nextTick(() => cb(null, true));
    });
    Role.registerResolver('Signed In Users', function(role, context, cb) {
      var userId = context.accessToken.userId;
      if (userId) return process.nextTick(() => cb(null, true));
      else return process.nextTick(() => cb(null, false));
    });

    // create resolver for all roles (Network Owner inherits all possible roles)
    let appRoles = buildAllRoles([], ['Network Owner']);
    _.forEach(appRoles, function(appRole) {
      // resolver for app roles
      Role.registerResolver(appRole, function(role, context, cb) {
        // verify login
        var userId = context.accessToken.userId;
        if (!userId) {
          // user is NOT logged in: callback with FALSE
          return process.nextTick(() => cb(null, false));
        }

        let req = context.remotingContext.req;
        let res = context.remotingContext.res;

        // default value for lookup
        let uroFor = 'Network', uroForId = 0;
        // determine what role is for
        if (appRole.startsWith('Site ')) {
          // make sure we have id of site been managed
          if (!res.locals || !res.locals.themeData || !res.locals.themeData.managingSite) {
            return process.nextTick(() => cb(null, false));
          }
          uroFor = 'Site', uroForId = res.locals.themeData.managingSite;
        }

        // check if user has role for the app
        getUserAppRoles(app, userId, uroFor, uroForId, (roles) => {
          if (roles && -1 !== roles.indexOf(appRole)) {
            // has role
            return process.nextTick(() => cb(null, true));
          }
          else {
            // tried all we can, but sorry no access
            return process.nextTick(() => cb(null, false));
          }
        });
      });
    });
  });
}

// get site roles for user
// TODO: save user's roles to cache for quicker request access
function getUserAppRoles(app, usrId, uroFor, uroForId, cb) {
  var zlUserRole = app.models.zlUserRole;
  let filter = {
    where: {usrId: usrId, uroFor: uroFor, uroForId: uroForId}
  };
  zlUserRole.findOne( filter, (err, data) => {
    if (data && data.uroRoles) {
      let allRoles = buildAllRoles([], data.uroRoles);
      cb(allRoles);
    }
    else if ('Network' !== uroFor) {
      // check if user has network role that inherits this role globally
      getUserAppRoles(app, usrId, 'Network', 0, cb);
    }
    else cb();
  });
}

// build a list of all roles and inherits
function buildAllRoles(all, roles) {
  all.push(...roles);
  _.forEach(roles, function(role) {
    // check if have inherits
    if (ROLE_INHERIT[role]) {
      all = buildAllRoles(all, ROLE_INHERIT[role]);
    }
  });
  return all;
}

module.exports = setup;
