const app = require('../server');
var nodemailer = require('nodemailer');
var request = require('request');

module.exports = function(app) {
  // provide functions
  app.locals.getInheritTheme = getInheritTheme;
  app.locals.cache = cache(app);
  app.locals.bypass = bypass;
  app.locals.authCtx = authCtx;
  app.locals.cdn = cdn;
  app.locals.mailer = nodemailer.createTransport(
    app.locals.config.mail.transport,
    app.locals.config.mail.defaults
  );
  app.locals.mailCommon = mailCommon;
  app.locals.recaptchaVerify = recaptchaVerify;
}

// get theme inheritance
function getInheritTheme(theme, tree = {}) {
  if (!app.locals.themes[theme]) {
    // this theme is not installed
    app.error('requested theme:%s but it is not installed', theme);
    return tree;
  }

  // add the one been requested
  tree[theme] = {};

  // get inheritance
  var config = app.locals.themes[theme].zealderConfig;
  if (config && config.themeDependency && !tree[config.themeDependency]) {
    tree = getInheritTheme(config.themeDependency, tree);
  }

  return tree;
}

// currently only redis cache store is supported
function cache(app) {
  if (app.locals.config.cacheDb && 'redis' === app.locals.config.cacheDb.connector) {
    // use redis for cache
    let redis = require("redis").createClient(app.locals.config.cacheDb);
    redis.on("error", function (err) {
      app.error(err);
    });
    // clear existing cache on startup
    redis.del('zlSites', 'themeTrees', 'themesConfig');

    return {
      // for getting data
      get: function () {
        if (3 == arguments.length) {
          // hash request
          redis.hget(arguments[0], arguments[1], arguments[2]);
        }
        else if (2 == arguments.length) {
          // key request
          redis.get(arguments[0], arguments[1]);
        }
      },
      // for setting data
      set: function () {
        if (4 == arguments.length) {
          // hash request
          redis.hset(arguments[0], arguments[1], arguments[2], arguments[3]);
        }
        else if (3 == arguments.length) {
          // key request
          redis.set(arguments[0], arguments[1], arguments[2]);
        }
      },
      // for setting volatile data with an expiration date
      setvol: function () {
        if (4 == arguments.length) {
          // key request
          redis.set(arguments[0], arguments[1], () => {
            redis.expire(arguments[0], arguments[2], arguments[3]);
          });
        }
      },
      // for deleting data
      del: function () {
        if (3 == arguments.length) {
          // hash request
          redis.hdel(arguments[0], arguments[1], arguments[2]);
        }
        else if (2 == arguments.length) {
          // key request
          redis.del(arguments[0], arguments[1]);
        }
      }
    }
  }
  else {
    // not using cache
    app.error('No config.cacheDb provided. This may affect performance.');
    let e = function() { arguments[arguments.length-1](); }
    return { get: e, set: e, setvol: e, del: e }
  }
}

// adds function to bypass certain condition of request
function bypass(req, conditions = {files: 1, api: 1}) {
  if (conditions.files) {
    // bypass all files request
    let regexp = /\w+\.\w+$/i;
    if (req.path.match(regexp)) {
      return true;
    }
  }
  if (conditions.api) {
    // bypass api request
    let regexp = /^\/api\//i;
    if (req.path.match(regexp)) {
      return true;
    }
  }
}

// for making context to pass to acl checker
function authCtx(req, res) {
  let ctx = {
    accessToken: null,
    remotingContext: { req: req, res: res }
  }
  // populate accessToken
  if (req && req.accessToken) ctx.accessToken = req.accessToken;

  return ctx;
}

// manage files saved locally
const cdn = {
  // function for upload
  upload: (forType, forId, name, data, params, cb) => {
    // TODO: make save in local 'content' folder
  }
}

// add common details to email message
function mailCommon(type, data, req, res) {
  const { zlSite } = res.locals;

  if ('form-notice' === type) {
    if (data.text) {
      data.text += `\nReferred By: ${req.cookies.referer}\n\n`+
        `-----------------------------------------------------------`+
        `\nThis message was triggered by a website user's request.\n`+
        `From IP: ${req.ip}\nFrom Url: ${req.get('Referrer')}\n\n`+
        `This message was sent from an ${process.env.ZEAL_NETWORK_NAME} powered `+
        `Website. Login to your account to view more details.`
    }
  }

  // make from name use site name (but auth email address so it will relay)
  data.from = `"${zlSite.zsName}" ${process.env.ZEAL_NETWORK_EMAIL}`;

  if (zlSite.zsInternalConfig && zlSite.zsInternalConfig.displayEmail) {
    data.replyTo = `"${zlSite.zsName}" ${zlSite.zsInternalConfig.displayEmail}`;
  }
  return data;
}

// to verify recaptcha
function recaptchaVerify(req, res, action = 'form_submit', cb, token = '', threshold = 0.3) {
  const { zlSite } = res.locals;
  if (zlSite.user) {
    // signed in user don't need to verify
    cb();
  }
  else {
    if (!token && 'POST' === req.method) {
      token = req.body.recaptcha;
    }

    if (!token) {
      cb('Token is required.');
    }
    else {
      request.post({
        url: 'https://www.google.com/recaptcha/api/siteverify',
        json: true,
        form: {
          secret: req.app.locals.reCAPTCHAsecret,
          response: token,
          remoteip: req.ip
        }
      }, function (error, response, body) {
        if (error) {
          req.app.error(error);
          // no error of user fault so allow action
          cb();
        }
        else if (body && body.success) {
          if (body.hostname !== req.hostname) {
            cb(`site: ${req.hostname} did not match recaptcha from ${body.hostname}`);
          }
          else if (body.score < threshold) cb('score below threshold');
          else if (action !== body.action) {
            cb(`action: ${action} did not match recaptcha action: ${body.action}`);
          }
          else cb();
        }
        else cb('unknown error occurred');
      });
    }
  }
}
