var path = require('path');
var loopback = require('loopback');
const { createLogger, format, transports } = require('winston');
var _ = require('lodash');
const { spawn } = require('child_process');
var fs = require('fs');

module.exports = function(app) {
  const { config } = app.locals;

  // ready function for debug logging
  app.debug = require('debug')('zealder:info');
  app.debug.log = console.info.bind(console);
  app.error = require('debug')('zealder:error');

  // ready function for activity logging
  app.logger = createLogger({
    format: format.combine(
      format.splat(),
      format.simple()
    ),
    transports: [
      new transports.Console({ format: format.simple() })
    ]
  });

  // prepare default themes and apps
  app.locals.defaultTheme = 'zealder-default-theme';
  app.locals.themes = {};
  app.locals.apps = {};
  app.locals.zlScripts = {};
  // id of admin site
  app.locals.adminSiteId = 2;
  if (process.env.ADMIN_ZSID) app.locals.adminSiteId = process.env.ADMIN_ZSID;

  // make easy access to certain env
  app.locals.reCAPTCHAkey = process.env.RECAPTCHA_KEY;
  app.locals.reCAPTCHAsecret = process.env.RECAPTCHA_SECRET;

  // provide easy means for packages to autoload models
  app.locals.loadModels = loadModels;

  // check if we should use cached for installed packages
  if (config.zlPacksUpdTime && config.cacheDb) {
    app.locals.cache.get('zlPacks', config.zlPacksUpdTime, (err, data) => {
      if (data) {
        app.debug(
          'Loading Zealder Packages list from cache. Any package installed or '
          +'removed since %s will not be reflected until value for '
          +'zlPacksUpdTime is changed.',
          config.zlPacksUpdTime
        );
        // load cached package info into local state
        setPackagesState(app, JSON.parse(data), null);
      }
      else {
        // delete all existing cache
        app.locals.cache.del('zlPacks', () => {});

        // call function to get all installed zealder apps/themes
        getPackages(app, (packs) => {
          // save to cache
          let packsJson =  JSON.stringify(packs);
          app.locals.cache.set('zlPacks', config.zlPacksUpdTime, packsJson, () => {});
        });
      }
    });
  }
  else getPackages(app);

  // placeholder to keep list of sites that has route already setup
  app.locals.sitesRouted = [];
  // placeholder to indicate if a bundle is in progress
  app.locals.bundling = false;

  // setup view folders
  app.set('views', path.join(__dirname, '../views'));
  app.set('view engine', 'pug');
};

// get all installed zealder apps/themes
function getPackages(app, cb = null) {
  let stdout = '';
  const ls = spawn('npm', ['ls', '--json', '--prod', '--depth=0', '--long']);
  ls.stdout.on('data', data => {
    stdout += data;
  });
  ls.stderr.on('data', data => {
    app.error(`stderr: ${data}`);
  });
  ls.on('close', (code) => {
    if (0 !== code) app.error('child process exited with code %s', code);
    app.debug('loading installed zealder packages');
    var p = JSON.parse(stdout);
    if (p.dependencies) {
      // now set packages to state
      setPackagesState(app, p.dependencies, cb);
    }
  });
}

// for saving packages info to state
function setPackagesState(app, dependencies, cb) {
  var zealPacks = {};
  let promises = [];

  _.forEach(dependencies, function(val, key) {
    buildPackagesList(app, zealPacks, val, key, promises);
  });
  /* app.debug('Installed themes: %O', app.locals.themes);
  app.debug('Installed apps: %O', app.locals.apps); */
  Promise.all(promises).then(() => {
    app.emit('zealder packages loaded');

    // if callback passed
    if (cb) cb(zealPacks);
  });
}

function buildPackagesList(app, zealPacks, val, key, promises) {
  const {config} = app.locals;

  if (zealPacks[key]) {
    // uhoh! looks like duplicate
    return app.debug('skipping duplicate package: %s', key);
  }

  // values that zealder need
  let value = {
    version: val.version,
    keywords: val.keywords,
    directories: val.directories,
    path: val.path,
    main: val.main,
    zealderConfig: val.zealderConfig
  }
  var isZealWare = 0;
  if ('zealder-cms' === key) {
    // load info about zealder-cms
    app.locals.zlVersion = value.version;
    app.locals.zlConfig = value;
    zealPacks[key] = value;
    // keep list of CMS dependencies so Webpack can use as alias
    value._dependencies = val._dependencies;
    app.locals.dependencies = value._dependencies;
    app.debug('running zealder-cms %s', value.version);
  }
  else if (value.keywords) {
    if (value.keywords.includes('zealder-theme')) {
      app.debug('loading theme %s', key);
      app.locals.themes[key] = value;
      isZealWare = 1;

      // check if we have servBundle
      let servBundle = `${path.join(process.cwd(), config.contentPath)}/build/${key}/server-render.js`;
      if ('development' !== process.env.NODE_ENV && !process.env.WORK
        && fs.existsSync(servBundle)) {
        app.locals.themes[key].servBundle = require(servBundle);
      }
    }
    if (value.keywords.includes('zealder-app')) {
      app.debug('loading app %s', key);
      app.locals.apps[key] = value;
      isZealWare = 1;
    }
    if (isZealWare) {
      zealPacks[key] = value;
      if (value.directories && value.directories.assets) {
        // make public asset folder for middleware
        let opt = null;
        if ('production' === app.get('env')) {
          // in production, use long-term cache
          opt = { immutable: true, maxAge: '10y'}
        }
        app.use(
          `/assets/${key}`,
          loopback.static(path.join(value.path, value.directories.assets), opt)
        );
      }
      if (value.main) {
        // check what package has and require them for quick use later
        let main = require(`${value.path}/${value.main}`);
        if (main.boot) {
          // package has own boot script
          app.debug('Bootstrapping %s', key);
          main.boot(app);
        }
        if (main.middleware) {
          // package has middleware
          value.middleware = main.middleware;
        }
        if (main.serverMount) {
          // package has serverMount
          value.serverMount = main.serverMount;
        }
      }
    }
  }
  // load scripts (if any)
  if (value.zealderConfig && value.zealderConfig.scripts) {
    _.forEach(value.zealderConfig.scripts, function(script, i) {
      app.locals.zlScripts[i] = `${value.path}/${script}`;
    });
  }
  // load models (if any)
  if (value.zealderConfig && value.zealderConfig.modelSources) {
    // mutate into list of scripts
    _.forEach(value.zealderConfig.modelSources, function(src, ds) {
      promises.push( new Promise((resolve) => {
        loadModels(app, ds, `${value.path}/${src}`, resolve);
      }));
    });
  }
  // now load subPackages (if any)
  if (value.zealderConfig && value.zealderConfig.subPackages) {
    _.forEach(value.zealderConfig.subPackages, function(subVal) {
      let sub = Object.assign(subVal, { path: `${value.path}/${subVal.subPath}`});
      buildPackagesList(app, zealPacks, sub, sub.name, promises);
    });
  }
}

// for loading models
function loadModels(app, datasource, src, cb = null) {
  const ds = app.dataSources[datasource];
  let promises = [], modelsNames = [];

  let files = fs.readdirSync(src);
  if (files) {
    files.forEach(file => {
      // load model for all model json files
      if (file.endsWith('.json')) {
        promises.push( new Promise((resolve, reject) => {
          fs.readFile(`${src}/${file}`, 'utf8', (fileErr, data) => {
            if (fileErr) app.error(`Failed reading %s because: %s`, file, fileErr);
            else {
              let mJson = JSON.parse(data);
              let model = loopback.createModel(mJson);
              app.model(model, {dataSource: datasource});
              modelsNames.push(mJson.name);
              resolve();
            }
          });
        }));
      }
    });
  }

  Promise.all(promises).then(() => {
    // process addl model js
    if (files) {
      files.forEach(file => {
        if (file.endsWith('.js')) {
          require(`${src}/${file}`)(app);
        }
      });
      /**
       * NOTE: UNCOMMENT TO UPDATE DB ON PRODUCTION ALTHOUGH THAT SHOULD BE
       * AVOIDED BECAUSE IT RECREATE TABLE STRUCTURE AND REMOVES INDEXES, while
       * TRYING TO REBUILD STRUCTURE/INDEX IT CAN LEAVE SOME OUT ON LARGE DB AND
       * CREATE OPTIMIZATION NIGHTMARE
       */
      //if ('build' === process.env.WORK && !_.isEmpty(modelsNames)) {
      if ('development' === process.env.NODE_ENV && !_.isEmpty(modelsNames)) {
        ds.autoupdate(modelsNames);
      }
    }
    if (cb) cb();
  });
}
