// @flow
'use strict';

var _ = require('lodash');
const EventEmitter = require('events');
var vhost = require('../helpers/vhost');
var path = require('path');
var loopback = require('loopback');

module.exports = function() {
  return function myMiddleware(req, res, next) {
    // determine which site/app is requested
    var app = req.app;
    const { config } = app.locals;

    if (config.proxyAuth) {
      // when forcing the use of proxy
      if (!req.get('X-Proxy-Auth') || req.get('X-Proxy-Auth') !== config.proxyAuth) {
        return res.status(401).send();
      }
    }

    // log request
    app.logger.info(`${req.hostname} ${req.ip} "${req.method} ${req.originalUrl}"`);

    let promises = [];

    promises.push( new Promise((resolve, reject) => {
      if ('zealder-builder' === req.get('User-Agent')) {
        // when request is from zealder-builder, we use X-Managing-Site
        if (req.get('X-Managing-Site')) {
          app.models.zlSite.findOne(
            {where: { zsId: req.get('X-Managing-Site')}}, (err, zs) => {
            if (err) app.zErr(err);

            // no site was found that match domain
            if (_.isEmpty(zs)) {
              app.error(
                'request to Edit site zsId:%s but it wasnt fount',
                req.get('X-Managing-Site'));
            }
            else {
              // to ensure it passes vhost test
              req.headers.host = zs.zsDomain;
              resolve([zs]);
            }
          });
        }
        else res.status(401).send('Managing Site is required');
      }
      else {
        // get zlSite from App State
        app.debug('checking if %s is in cache', req.hostname);
        app.locals.cache.get('zlSites', req.hostname, (err, data) => {
          if (data) {
            app.debug('loading %s from cache', req.hostname);
            resolve(JSON.parse(data));
          }
          else {
            app.debug('fetching sites for %s from model', req.hostname);
            app.models.zlSite.find(
              {where: { zsDomain: req.hostname}}, (err, zlSites) => {
              if (err) app.zErr(err);

              // no site was found that match domain
              if (_.isEmpty(zlSites)) {
                app.debug('no sites found for %s in model', req.hostname);
                return next();
              }

              // make site based on url path
              let domainSites = [];
              _.forEach(zlSites, function(val) {
                val = _.cloneDeep(val.toJSON());
                if (!app.locals.themes[val.zsTheme]) {
                  // since site has no theme, use default theme
                  val.zsTheme = app.locals.defaultTheme;
                }
                domainSites.push(val);
              });

              // setup zlSite in App State
              let dsites =  JSON.stringify(domainSites);
              app.locals.cache.set('zlSites', req.hostname, dsites, () => {});
              resolve(domainSites);
            });
          }
        });
      }

    }));

    Promise.all(promises).then(val => {
      val = val[0];
      let site, skipCb = false;;
      if (1 < val.length) {
        // determine which site is requested when more than one site shares a domain
        let path = req.path;
        // api call will pass as header instead of URL path
        if (req.get('X-Base-Url-Path')) path = '/'+req.get('X-Base-Url-Path')+'/';
        site = _.find(val, function(o) {
          if (path == '/'+o.zsBaseUrlPath) {
            // missing ending slash, redirect
            let query = '';
            if (req.query) {
              _.forEach(req.query, function(v, key) {
                query = (query)? `${query}&${key}=${v}`: `?${key}=${v}`;
              });
            }
            skipCb = true;
            return res.redirect('/'+o.zsBaseUrlPath+'/'+query);
          }
          return _.startsWith(path, o.zsBaseUrlPath+'/', 1);
        });
        if (!site) {
          // couldn't find sub-dir site so it must be root site
          site = _.find(val, function(o) {
            return (o.zsBaseUrlPath)?false:true;
          });
        }
      }
      else site = val[0];


      res.locals.zlSite = _.cloneDeep(site);
      // local path where site content will save
      res.locals.zlSite.contentPath = `${app.locals.config.contentPath}/site/${res.locals.zlSite.zsId}`;

      // setup other values we will need at some point
      res.locals.htmlClass = [];
      res.locals.themeData = {};
      res.locals.app_scripts = {},
      res.locals.app_stylesheets = {};
      res.locals.appStyles = {};
      res.locals.app_headcode = {};
      res.locals.app_bodycode = {};
      res.locals.socketQuery = { zsId: res.locals.zlSite.zsId };
      res.locals.zlSite.siteContents = {};
      class MyEmitter extends EventEmitter {};
      res.locals.emiter = new MyEmitter();

      if (!app.locals.sitesRouted[res.locals.zlSite.zsId]) {
        // make public folder for site (in case not using cloud/cdn)
        let siteRoot = (res.locals.zlSite.zsBaseUrlPath)?`/${res.locals.zlSite.zsBaseUrlPath}/`:`/`;
        let opt = null;
        if ('production' === app.get('env')) {
          // in production, use long-term cache
          opt = { immutable: true, maxAge: '10y'}
        }
        app.use(siteRoot, vhost(
            res.locals.zlSite.zsDomain, 'IS_ROUTE',
            loopback.static(path.join(process.cwd(), res.locals.zlSite.contentPath, 'dist'), opt)
          )
        );
        app.locals.sitesRouted[res.locals.zlSite.zsId] = true;
      }

      if (!skipCb) next();
    });
  }
};
