//@flow

'use strict';

var _ = require('lodash');
var app = require('../server');
var events = require('../../socket/actions');
var sanitize = require('../helpers/sanitize');

module.exports = function(Zlsitecontentset) {
  // method for reordering a contentset
  Zlsitecontentset.changeorder = function(scsId, order, cb) {
    // get info about current set
    Zlsitecontentset.findById(scsId, (err, scs) => {
        // get all sets at same level
        Zlsitecontentset.find(
          {
            order: "scsOrder ASC",
            where: {scsParent: scs.scsParent, zlSiteId: scs.zlSiteId}
          }, (err, data) => {
            let promises = [Promise.resolve()];

            // total items/highest order
            let total = data.length;
            // fix bad values
            if (order < 1) {
                order = 1;
            }
            else if (order > total) {
                order = total;
            }
            // set new order values
            let key = 0;
            for (let i = 1; i <= total; i++) {
              promises.push( new Promise((resolve, reject) => {
                if (i == order) {
                    // force the new value
                    Zlsitecontentset.updateAll(
                      {scsId: scsId},
                      {scsOrder: i, scsModified: scs.scsModified},
                      function(err, info) {
                        if (!err) {
                          resolve();
                        }
                      });
                }
                else if (data[key]['scsId'] == scsId) {
                    // don't update the based item since it must force new order
                    key++;
                    i--; // keep current order
                    resolve();
                }
                else {
                    // all others keep existing location relevance
                    Zlsitecontentset.updateAll(
                      {scsId: data[key]['scsId']},
                      {scsOrder: i, scsModified: data[key]['scsModified']},
                      function(err, info) {
                        if (!err) {
                          resolve();
                        }
                      });
                    key++;
                }
              }));
            }

            Promise.all(promises).then(values => {
              refreshSiteMenu(scs.zlSiteId);
              // all promises settled so run callback
              cb(null, values);
            });
          });
      });
  }
  // method for getting a site's menu pages
  Zlsitecontentset.menupages = function(zsId, req, res, cb) {
    Zlsitecontentset.find(
      {
        where: {zlSiteId: zsId },
        fields: { }, order: ['scsParent ASC', 'scsOrder ASC']
      }, (err, zlSitePages) => {
        let promises = [];
        let pages = makeEntity(req, res, zlSitePages, 0, {}, '', promises);
        Promise.all(promises).then(() => {
          // start new promise for any other functions that need to modify pages
          // before we callback
          let cbPromises = [];
          res.locals.emiter.emit('pages-loaded', pages, cbPromises);
          Promise.all(cbPromises).then(() => {
            cb(null, pages);
          });
        });
    });
  }
  // method for getting page tree for select input type
  Zlsitecontentset.pageTreeSelect = function(zsId, forPage, cb) {
    // get pages for site
    let filter = {
      where: {scsType: 'page', zlSiteId: zsId},
      order: ['scsParent ASC', 'scsOrder ASC']
    }
    Zlsitecontentset.find(filter, (err, pages) => {
      let pageList = buildSelect(pages);
      cb(null, pageList);
    });

    var buildSelect = function (all, top = 0, data = [''], indent = '') {
      _.forEach(_.filter(all, ['scsParent', top]), function(val) {
        val = val.toJSON();
        data.push({[val.scsId]: `${indent} ${val.scsName}`});
        // check if have children
        if (_.find(all, ['scsParent', val.scsId])) {
          data = buildSelect(all, val.scsId, data, `${indent}-`);
        }
      });
      return data;
    }
  }

  // checks if a slug is already in use
  Zlsitecontentset.doesSlugExist = function(slug, data, cb) {
    // check in site
    Zlsitecontentset.findOne(
      {
        fields: {scsType: true},
        where: {scsSlug: slug, zlSiteId: data.zlSiteId, scsId: {neq: data.scsId}}
      }, (err, data) => {
        if (err) cb(null, {error: 'system error: cannot validate slug right now'});
        else if (data) {
          cb(null, {error: `This slug is already used by a ${data.scsType}. `
                        +'Please choose a different slug.'});
        }
        else cb(null, {});
      })
  }

  Zlsitecontentset.remoteMethod(
    'changeorder', {
      http: { path: '/changeorder', verb: 'post' },
      accepts: [
        {arg: 'id', type: 'number', required: true},
        {arg: 'order', type: 'number', required: true}
      ],
      returns: { arg: 'data', type: 'object' }
    }
  );

  Zlsitecontentset.remoteMethod(
    'menupages', {
      http: { path: '/menupages', verb: 'get' },
      accepts: [
        {arg: 'zsId', type: 'number', required: true, 'http': {source: 'query'}},
        {arg: 'req', type: 'object', 'http': {source: 'req'}},
        {arg: 'res', type: 'object', 'http': {source: 'res'}}
      ],
      returns: { arg: 'data', type: 'object' }
    }
  );

  Zlsitecontentset.remoteMethod(
    'doesSlugExist', {
      http: { path: '/doesSlugExist', verb: 'post' },
      accepts: [
        {arg: 'val', type: 'string', required: true},
        {arg: 'data', type: 'object', required: true}
      ],
      returns: { arg: 'data', type: 'object' }
    }
  );

  // operation hook when deleted
  Zlsitecontentset.observe('before delete', (ctx, next) => {
    // get site id
    Zlsitecontentset.findById(ctx.where.scsId, (err, scs) => {
      if ('page' == scs.scsType) {
        // for page, delete actual content since it wont be use anywhere else
        scs.zlSiteContents({ fields: { scId: true } }, (err, sc) => {
          _.forEach(sc, function(content) {
            content.destroy(()=>{});
          });
        });
      }
      // delete all content relationships for page/blogs/etc
      scs.zlSiteContents.destroyAll(()=>{});

      // get all child pages
      let filter = {
        where: {scsParent: scs.scsId, zlSiteId: scs.zlSiteId},
        fields: { scsId: true }
      }
      Zlsitecontentset.find(filter, (err, scsSubs) => {
        _.forEach(scsSubs, function(scsSub) {
          scsSub = scsSub.toJSON();
          // delete child page
          Zlsitecontentset.destroyAll({ scsId: scsSub.scsId, noRfrsh: 1}, () => { });
        });
        if (ctx.where.noRfrsh) delete ctx.where.noRfrsh;
        else refreshSiteMenu(scs.zlSiteId);
        next();
      });
    });
  });

  // operation hook for before saving data
  Zlsitecontentset.observe('before save', (ctx, next) => {
    if (ctx.instance && ctx.isNewInstance) {
      const scs = ctx.instance;
      // make sure there is an order
      if (!scs.scsOrder) {
        scs.scsOrder = 1;
        // get value for next "order"
        let filter = {
          order: "scsOrder DESC", limit: 1,
          where: {scsParent: scs.scsParent, zlSiteId: scs.zlSiteId},
          fields: { scsOrder: true }
        }
        Zlsitecontentset.find(filter, (err, data) => {
          if (!_.isEmpty(data) && data[0].scsOrder) {
            scs.scsOrder = data[0].scsOrder + 1;
          }
          next();
        });
      }
      // make sure we have a scsSlug
      if (!scs.scsSlug) {
        scs.scsSlug = sanitize.slugify(scs.scsName);
      }
      else next();
    }
    else next();
  });

  // operation hook for after saving data
  Zlsitecontentset.observe('after save', (ctx, next) => {
    if (ctx.instance) {
      refreshSiteMenu(ctx.instance.zlSiteId);
    }
    next();
  });
};

// makes list of pages for main or subsite
function makeEntity(req, res, all, top, data, path = '', promises) {
  var Role = app.models.Role;
  const { zlSite } = res.locals;
  const zsBaseUrlPath = (zlSite.zsBaseUrlPath) ? `/${zlSite.zsBaseUrlPath}/` : '/'

  _.forEach(_.filter(all, ['scsParent', top]), function(val) {
    val = val.toJSON();

    /* check if have children
     NOTE: since children loads first, they will return even if parent
     permission is denied, so must assign children correct permission/role */
    if (_.find(all, ['scsParent', val.scsId])) {
      data = makeEntity(req, res, all, val.scsId, data, path+val.scsSlug+'/', promises);
    }

    let loadThisEntity = () => {
      data[val.scsId] = val;
      // placeholder for content id list
      data[val.scsId]['contents'] = [];
      // set page path
      data[val.scsId]['path'] = `${path}${val.scsSlug}`;
      data[val.scsId]['realPath'] = `${zsBaseUrlPath}${data[val.scsId]['path']}`;
    }

    // make sure they have correct permission
    // verify that they have permission to this page
    let ctx = app.locals.authCtx(req, res);
    return promises.push( new Promise((resolve, reject) => {
      Role.isInRole(val.scsPermission, ctx, (err, isInRole) => {
        if (isInRole) {
          loadThisEntity();
        }
        resolve();
      });
    }));

    // if we reach here, I guess we can load page
    loadThisEntity();
  });

  return data;
}

function refreshSiteMenu(zsId) {
  // broadcast update to all users connected to the updated site
  app.io.in('site-'+zsId).emit(events.MENU_UPDATED);
}
