//@flow
'use strict';

var _ = require('lodash');
const app = require('../../server/server');
const events = require('./../../socket/actions');

function setup(zlUser) {
  // operation hook for after saving data
  zlUser.observe('after save', afterSave);

  // remote methods
  // before login
  zlUser.beforeRemote('login', beforeLogin);
  // before create
  zlUser.beforeRemote('create', beforeCreate);
  // after create
  zlUser.afterRemote('create', afterCreate);
  // before reset
  zlUser.beforeRemote('resetPassword', beforeResetPassword);

  // events
  // on resetPasswordRequest
  zlUser.on('resetPasswordRequest', resetPasswordRequest);

  // remote methods
  _.forEach(remoteMethods, function(val, key) {
    zlUser.remoteMethod( key, val );
  });
  zlUser.userFromToken = userFromToken;
  zlUser.simulateUser = simulateUser;

  /* Delete built in validations from email and username properties
  to allow use of same email address in different realms
  @see: https://github.com/strongloop/loopback/issues/1987#issuecomment-270357932 */
  delete zlUser.validations.email;
  delete zlUser.validations.username;
  // Adds email format validation
  // Note custom validator because validatesFormat with regex will return false on a null value
  var re = /^(([^<>()[\]\\.,;:\s@\"]+(\.[^<>()[\]\\.,;:\s@\"]+)*)|(\".+\"))@((\[[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\])|(([a-zA-Z\-0-9]+\.)+[a-zA-Z]{2,}))$/;
  zlUser.validate('email', function (err) {
    if (!re.test(this.email) && this.email !== undefined) err();
  }, {message: 'Email format is invalid'});
}
module.exports = setup;

function afterSave(ctx, next) {
  if (ctx.instance && app.io) {
    let usrId = ctx.instance.usrId;
    // broadcast update to all site that may be displaying user profile
    app.io.in('usesProfile-'+usrId).emit(events.USER_UPDATED, usrId);
  }
  next();
}

function beforeLogin(ctx, inst, next) {
  const { zlSite } = ctx.res.locals;
  _addRealmToCtx(zlSite, ctx.args.credentials);
  /* we don't use recaptcha for login because it could be from other source
  (e.g., 3rd party app) */
  next();
}

function beforeCreate(ctx, inst, next) {
  const { zlSite } = ctx.res.locals;
  _addRealmToCtx(zlSite, ctx.args.data);
  ctx.req.app.locals.recaptchaVerify(ctx.req, ctx.res, 'signup', (err) => {
    if (err) {
      // pretend form was submitted
      ctx.res.status(401).send('Login failed');
    }
    else { next(); }
  });
}

function afterCreate(ctx, inst, next) {
  // login user
  inst.createAccessToken(null, (err, token) => {
    if (err) app.error(err);
    else ctx.result = token;
    next();
  });
}

function beforeResetPassword(ctx, inst, next) {
  const { zlSite } = ctx.res.locals;
  _addRealmToCtx(zlSite, ctx.args.options);
  ctx.args.options = Object.assign(ctx.args.options, {res: ctx.res, req: ctx.req});
  next();
}

function resetPasswordRequest(info) {
  // send email with instructions to reset password
  const { zlSite } = info.options.res.locals;
  var url = `${zlSite.zsProtocol}://${zlSite.zsDomain}?initModal=reset-password`+
    `&access_token=${info.accessToken.id}`;
  let data = {
      to: info.email,
      subject: 'Password reset',
      html: '<p>Hello,</p><p>This is a response to your request to recover your'+
            ' password.</p><p>To reset your password click the link below, or copy '+
            'and paste the link into the address bar of your browser:</p><a href="'
            + url + '">'+url+'</a></p><p><b>THIS LINK WILL EXPIRE IN 15 MINUTES.'+
            '</b><p>If you did not make this request, please ignore this '+
            'message.</p><p>Thank you,<p>' + zlSite.zsName
  };
  data = app.locals.mailCommon('transactional', data, info.options.req, info.options.res);
  app.locals.mailer.sendMail(data, (err, info) => {
    if (err) app.error(err);
    else app.debug(`Password reset email sent to ${data.to}`)
  });
}

function _addRealmToCtx(zlSite, data) {
  // add realm
  if (zlSite && data) {
    let realm;
    if (zlSite.zsOptions && zlSite.zsOptions.realm) {
      realm = zlSite.zsOptions.realm;
    }
    else realm = zlSite.zsId;

    data.realm = realm;
  }
}

const remoteMethods = {
  'userFromToken': {
    description: 'Get user detail from accessToken',
    http: { path: '/userFromToken', verb: 'get' },
    accepts: [
      { arg: 'req', type: 'object', 'http': {source: 'req' }}
    ],
    returns: { arg: 'data', type: 'object' }
  },
  'simulateUser': {
    description: 'Allows Network Admin to simulate a user by returning token',
    http: { path: '/simulateUser', verb: 'get' },
    accepts: [
      { arg: 'username', type: 'any', required: true, 'http': {source: 'query'}},
      { arg: 'req', type: 'object', 'http': {source: 'req' }}
    ],
    returns: { arg: 'data', type: 'object' }
  }
}

// Get user detail from accessToken
function userFromToken(req, cb) {
  const { accessToken } = req;

  app.models.zlUser.findById(accessToken.userId, (err, usr) => {
    if (!err && usr) {
      cb(null, usr);
    }
    else cb(err);
  });
}

function simulateUser(username, req, cb) {
  app.models.zlUser.findOne({where: {username: username}}, (err, usr) => {
    if (err) cb(err);
    else if (usr) {
      usr.createAccessToken(3600, (err, token) => {
        if (err) app.error(err);
        cb(null, token);
      });
    }
    else cb('user does not exist');
  });
}
