// @flow
'use strict';

var async = require('async');
var _ = require('lodash');
var sanitize = require('../../../../server/helpers/sanitize');

function middleware(req, res, next) {
  const {zlSite, themeData} = res.locals;

  // never allow robot for admin
  res.locals.robots = 'NOINDEX, NOFOLLOW';

  // make sure they are signed in
  if (!zlSite.user || !zlSite.user.usrId) {
    if (!req.xhr && zlSite.zsOptions && zlSite.zsOptions.loginUrl) {
      // send them to login page for this site
      res.redirect(zlSite.zsOptions.loginUrl);
    }
    else res.status(401).send('Login required to access this page');
  }
  else {
    // wp
    if (zlSite.zsOptions && zlSite.zsOptions.wpUrl) {
      themeData.wpUrl = zlSite.zsOptions.wpUrl;
    }

    async.series([
      (cb) => require('./get-auth-sites')(req, res, cb),
      (cb) => siteAppsAdmin(req, res, cb)
    ], (err, results) => {
      // if this theme is a denpendency of another and other theme listen to
      // this event, it is responsible for calling next
      if (!res.locals.emiter.emit('zealder-admin-theme loaded', next)) {
        // since there were no listeners, we can safely call next
        next();
      }
    });
  }
}
module.exports = middleware;

// load admin functions for site apps
function siteAppsAdmin(req, res, cb) {
  var { themeData } = res.locals;
  const { apps, getInheritTheme, themes } = req.app.locals;

  if (!themeData.managingSite) cb();
  else {
    // provide function for apps/themes to add pages
    res.locals.addPages = function(appPages) {
      addPages(req, res, appPages);
    }

    let mgSiteApps = themeData.authSites[themeData.managingSite].zsApps;
    if (mgSiteApps) {
      req.app.debug('Site has apps %o', mgSiteApps);

      // iterate managingSite apps
      _.forEach(mgSiteApps, function(key) {
        if (apps[key] && apps[key].main) {
          let main = require(`${apps[key].path}/${apps[key].main}`);

          // check if middleware for admin is provided
          if (main.admin) {
            main.admin(req, res);
          }
        }
      });
    }

    let mgSiteTheme = themeData.authSites[themeData.managingSite].zsTheme;
    if (mgSiteTheme) {
      let themeTree = getInheritTheme(mgSiteTheme);
      // iterate managingSite themetree
      _.forEach(themeTree, function(v, key) {
        if (themes[key] && themes[key].main) {
          let main = require(`${themes[key].path}/${themes[key].main}`);

          // check if middleware for admin is provided
          if (main.admin) {
            main.admin(req, res);
          }
        }
      });
    }

    cb();
  }
}

function addPages(req, res, appPages) {
  var Role = req.app.models.Role;
  if (_.isArray(appPages)) {
    let ctx = req.app.locals.authCtx(req, res);
    res.locals.emiter.on("pages-loaded", function(pages, promises) {
      req.app.debug('Adding Admin pages %o', appPages);
      promises.push( new Promise(function(resolve, reject) {
        let pgKey = 0;
        async.eachSeries(appPages, function(page, callback) {
          // make sure required fields were provided
          if (!page.name || !page.id) {
            return req.app.error('missing "name" and/or "id" prop in: %o', page);
          }
          // add page while merge with defaults for app pages
          pgKey = pgKey + 1;
          let pg = {
            "scsId": page.id,
            scsName: page.name,
            "scsType": "page",
            "scsStatus": "active",
            "scsOrder": (page.order)?page.order:1000,
            "scsSlug": page.slug,
            "scsParent": 0,
            "scsHidden": (page.hidden)?true:false,
            "scsBase": "standard",
            "scsPermission": (page.role)?page.role:"Signed In Users",
            "scsOptions": {
              UniId: page.id,
              iconBundle: (page.iconBundle)?page.iconBundle:'fontello',
              iconGlyph: (page.iconGlyph)?page.iconGlyph:''
            },
            "zlSiteId": res.locals.zlSite.zsId,
            "path": "",
            "contents": []
          };
          // make sure we have auth for this page
          Role.isInRole(pg.scsPermission, ctx, function(err, isInRole) {
            if (isInRole) {
              if (!pg.scsSlug) {
                // no slug, no sweat. Let's make one
                pg.scsSlug = sanitize.slugify(pg.scsName);
              }
              if (page.parentId) {
                // assign a parent by searching parent's UniId
                let parent = _.find(pages, function(o) {
                  return o.scsOptions && page.parentId == o.scsOptions.UniId;
                });
                if (parent && parent.scsId) {
                  pg.scsParent = parent.scsId;
                  pg.path = `${parent.path}/`;
                }
              }
              // add path
              // TODO: make sure path doesn't exist
              pg.path += pg.scsSlug;
              // now save to menu
              pages[page.id] = pg;
            }
            callback();
          });
        }, function(err) {
          resolve();
        });
      }));
    });

    /* listen for request on content
    NOTE: only for basic content. Page that use async/model to load contents
    should add own listener for content
    */
    _.forEach(appPages, function(page) {
      // attach contents to page
      if (page.content && _.isArray(page.content)) {
        // listen for request for content for this page
        res.locals.emiter.on(`fetching content ${page.id}`, (cb) => {
          // check permission
          let role = (page.role)?page.role:"Signed In Users";
          Role.isInRole(role, ctx, function(err, isInRole) {
            if (isInRole) {
              req.app.debug('Responding with "content" for %s', page.id);
              var data = [];
              _.forEach(page.content, function(content, contKey) {
                let contId = content.scId ? content.scId : `${page.id}-${contKey}`;
                let cont = {
                  scId: contId,
                  scTitle: content.scTitle ? content.scTitle : '',
                  scType: content.scType ? content.scType : 'Html',
                  scPlacement: 'body',
                  scContent: content.scContent ? content.scContent : ''
                }
                data.push(cont);
              });
              cb(null, data);
            }
            else cb('Page Authorization failed');
          });
        });
      }
    });
  }
}
