// @flow

import React from 'react';
import { FormGroup, ControlLabel, Checkbox, Radio, HelpBlock,
InputGroup, FormControl, OverlayTrigger, Tooltip, Popover } from 'react-bootstrap';
var _ = require('lodash');
import revalidator from 'revalidator';
import { connect } from 'react-redux';

class FormBuilderContainer extends React.Component {
  constructor(props) {
    super(props);

    let state = {
      validationStates: {},
      currentTT: '',
      currentErr: '',
      formData: props.values ? props.values : {}
    };
    // make sure we declare field so it won't be set as "uncontrolled" initially
    _.forEach(this.props.fields, field => {
      state.formData[field.name] = (field.value)?field.value:'';
    });
    this.state = state;

    this.fieldOverlay = {};

    this.handleChange = this.handleChange.bind(this);
    this.handleBlur = this.handleBlur.bind(this);
    this.handleSave = this.handleSave.bind(this);
    this.showValidationOverlay = this.showValidationOverlay.bind(this);
    this.ajaxValidate = this.ajaxValidate.bind(this);
    this.makeField = this.makeField.bind(this);
    this.validateField = this.validateField.bind(this);
  }

  handleChange(val='', field, promises) {
    // new state
    if (_.isObject(val)) {
      // when val is object (multi-selects, checkbox, etc) merge with existing
      val = Object.assign({}, this.state.formData[field.name], val);
    }
    let obj = {}; obj[field.name] = val;
    let formData = Object.assign({}, this.state.formData, obj);

    // hide any existing error
    if (this.state.currentTT) this.state.currentTT.hide();
    let validationStates = Object.assign({}, this.state.validationStates, {
      [field.name]: null
    })

    // update value to what was typed/provided
    this.setState({formData: formData, validationStates: validationStates});
  }

  validateField(val='', field, promises) {
    // default validationState
    let validationState = null;
    // hide any existing error
    if (this.state.currentTT) this.state.currentTT.hide();

    // validate field if required
    if (!_.isEmpty(field.validations)) {
      let validation = revalidator.validate(
        {[val]: val}, {properties: {[val]: field.validations}}
      );
      if (validation.valid) {
        if (val) {
          validationState = 'success';

          // when a field is for confirming another field value (eg, confirm password)
          if (field.validations.confirmationOf) {
            if (val !== this.state.formData[field.validations.confirmationOf]) {
              validationState = 'error';
              this.showValidationOverlay('Value did not match', field.name);
            }
          }

          // check for api validation
          if (field.validations.api) {
            let api = field.validations.api;
            // assign as warning until we complete validation
            validationState = 'warning';
            let msg = 'validating...';
            if (api.message) msg = api.message;
            this.showValidationOverlay(msg, field.name);
            // when list of promises exist we assign new
            if (promises) {
              promises.push( new Promise((resolve, reject) => {
                this.ajaxValidate(api, val, field.name, resolve);
              }));
            }
            else this.ajaxValidate(api, val, field.name);
          }
        }
      }
      else {
        validationState = 'error';
        this.showValidationOverlay(validation.errors[0].message, field.name);
      }
    }
    // reset validationState
    this.setState({
      validationStates: Object.assign({}, this.state.validationStates, {
      [field.name]: validationState
    })});
    return validationState;
  }

  handleBlur(val='', field) {
    // get current state and trim val
    val = val.trim();
    let formData = Object.assign({}, this.state.formData, {[field.name]: val});

    // TODO: run any passed onblur listeners
    if (this.props.onblur) {
      this.props.onblur();
    }
    // update value to what was typed/provided
    this.setState({formData: formData}, () => this.validateField(val, field));
  }

  handleSave() {
    let self = this, data = {}, promises = [], recaptcha = '';

    // validate all fields
    _.forEach(this.props.fields, function(field) {
      let validation = self.validateField(self.state.formData[field.name],
        field, promises);
      /* TODO: this may not really be async because 'validation' value may not
      be set on time */
      if ('error' == validation) {
        $(`#field-${field.name}`).focus();
        data = {};
        return false;
      }
      else {data[field.name] = self.state.formData[field.name];}
    });

    if (!this.props.user || !this.props.user.usrId) {
      promises.push( new Promise((resolve) => {
        // validate with recaptcha for unknown user
        let greProps = {action: 'form_submit'};
        if (this.props.formAction) {
          greProps.action = this.props.formAction;
        }
        grecaptcha.execute(reCAPTCHAkey, greProps).then((token) => {
          /* assign it to var instead of directly to 'data', cuz if 'data' has
          vale it will post even if there's error */
          recaptcha = token;
          resolve();
        });
      }));
    }

    Promise.all(promises).then(() => {
      // all promises settled so submit
      if (!_.isEmpty(data)) {
        if (recaptcha) { data['recaptcha'] = recaptcha; }
        if (this.props.handleSubmit) {
          this.props.handleSubmit(data, () => {

          });
        }
      }
    });
  }

  showValidationOverlay(msg, name) {
    this.fieldOverlay[name].show();
    this.setState({
      currentErr: msg,
      currentTT: this.fieldOverlay[name]
    });
  }

  ajaxValidate(api, val, name, resolve) {
    $.post(api, {val: val, data: this.state.formData})
     .done(response => {
       let validationState = 'success';
       if (this.state.currentTT) this.state.currentTT.hide();
       if (response.data.error) {
         validationState = 'error';
         this.showValidationOverlay(response.data.error, name);
       }
       else {
         if (resolve) resolve();
       }
       // reset validationState
       this.setState({
         validationStates: Object.assign({}, this.state.validationStates, {
         [name]: validationState
       })});
     })
  }

  makeField(fields, field, tooltip, FieldRenderer) {
    const { autoFocus } = this.props;
    // current value
    let val = this.state.formData[field.name];

    // for handling change in field
    let changeHdlr = (e)=> {
      let value = e.target.value;
      if (e.target.type === 'checkbox') {
        // for checkbox, we pass obj array instead
        value = { [e.target.value]: e.target.checked }
        // except boolean field
        if (field.boolean) value = e.target.checked;
      }
      this.handleChange(value, field);
    }

    // handle blur (exit field)
    let blurHdlr = (e)=> this.handleBlur(e.target.value, field);

    if (field.contigent) {
      // for field that are contigent on another field value, check requirments
      let failContigent;
      _.forEach(field.contigent, (contData, contField) => {
        let contVal = this.state.formData[contField];
        // check if val is in required list
        if (contData.in && -1 == contData.in.indexOf(contVal)) {
          failContigent = true;
          return false;
        }
        // check if val is in prohibited list
        if (contData.notin && -1 !== contData.notin.indexOf(contVal)) {
          failContigent = true;
          return false;
        }
      });
      if (failContigent) return fields;
    }

    if (field.hidden) {
      // for hidden fields
      fields.push(
        <input key={`col-${field.name}`} type="hidden" value={val}
          name={`field-${field.name}`} />
      );
      return fields;
    }
    // build controls
    var control = [];
    if ('checkbox' == field.control) {
      // Checkbox control
      _.forEach(field.options, function(option, key) {
        control.push(
          <Checkbox key={key} checked={val && (field.boolean || val[key])}
            onChange={changeHdlr} value={key} inline>
            {option}
          </Checkbox>
        );
      });
    }
    else if ('radio' == field.control) {
      // Radio control
      _.forEach(field.options, function(option, key) {
        control.push(
          <Radio key={key} name={field.name} checked={val == key} inline
            onChange={changeHdlr} value={key}>
            {option}
          </Radio>
        );
      });
    }
    else if ('select' == field.control) {
      // select control
      let opt = [];
      _.forEach(field.options, function(optItem, index) {
        // default is to use string as display and post value
        let value = optItem, displayValue = optItem;
        if (_.isPlainObject(optItem)) {
          // when value is object, then post value is first item
          value = _.findKey(optItem, () => true);
          displayValue = optItem[value];
        }
        opt.push(
          <option key={index} value={value}>{displayValue}</option>
        );
      });
      control = (
        <FormControl name={field.name} componentClass="select"
          placeholder={field.placeholder} value={val} onChange={changeHdlr}>
          {opt}
        </FormControl>
      );
    }
    else if ('textarea' == field.control) {
      // textarea control
      control = (
      <FormControl name={field.name}
        autoFocus={(1 > fields.length && autoFocus)?true:false}
        componentClass="textarea" placeholder={field.placeholder} value={val}
        onChange={changeHdlr} onBlur={blurHdlr} rows={field.rows?field.rows:2} />
      );
    }
    else {
      // input control
      control = (
      <FormControl name={field.name}
        autoFocus={(1 > fields.length && autoFocus)?true:false}
        type={field.format} placeholder={field.placeholder} value={val}
        onChange={changeHdlr} onBlur={blurHdlr} />
      );
      if (field.addons) {
        let addOnBefore = [], addOnAfter = [];
        _.forEach(field.addons, function(addon, index) {
          if (addon.before) {
            addOnBefore.push(
              <InputGroup.Addon key={index}>{addon.before}</InputGroup.Addon>
            );
          }
        });
        control = (
          <InputGroup>
            {addOnBefore && addOnBefore}
            {control}
          </InputGroup>
        );
      }
    }
    // prepare how to display help info
    let help = (_.isArray(field.help))?field.help.join(''):field.help;

    let labelClass;
    if ('-' === field.label) {
      // when a dash is used, make label invisible
      labelClass = 'invisible';
    }

    let formGroupClass;
    if (field.validations && false === field.validations.allowEmpty) {
      formGroupClass = 'required';
    }

    // add field
    let returnVal = (
      <FormGroup key={`field-${field.name}`} controlId={`field-${field.name}`}
        validationState={this.state.validationStates[field.name]}
        className={formGroupClass}>
        {field.label &&
          <ControlLabel>
            <span className={labelClass}>{field.label}</span>
            {help && (60 < help.length) &&
              <OverlayTrigger placement="bottom"
                overlay={
                  <Popover id={`popHelp-${field.name}`}>
                    <div dangerouslySetInnerHTML={{__html: help}} />
                  </Popover>
                }>
                <span style={{marginLeft: 5}}
                  className="icon-fontello-info-circle" />
              </OverlayTrigger>
            }
          </ControlLabel>
        }
        <OverlayTrigger placement="top" overlay={tooltip}
          ref={i => this.fieldOverlay[field.name] = i} trigger='click'>
          <div></div>
        </OverlayTrigger>
        {control}
        <FormControl.Feedback />
        {help && (60 >= help.length) &&
          <HelpBlock>{field.help}</HelpBlock>
        }
      </FormGroup>
    );

    // when a custom FieldRenderer is supplied, use it
    if (FieldRenderer) {
      fields.push(
        <FieldRenderer key={`col-${field.name}`} field={field}>
          {returnVal}
        </FieldRenderer>
      );
    }
    // render field
    else fields.push(returnVal);

    return fields;
  }

  componentWillReceiveProps(props) {
    // when caller updates default values
    if (props.values &&
      JSON.stringify(props.values) !== JSON.stringify(this.state.formData)) {
      // update state
      this.setState({formData: props.values});
    }
  }

  componentDidMount() {
    // listen for click on submitButton
    if (this.props.submitButton) {
      $(this.props.submitButton).on( "click", this.handleSave);
    }

  }

  componentWillUnmount(){
    // stop listen for click on submitButton
    if (this.props.submitButton) {
      $(this.props.submitButton).off( "click", this.handleSave);
    }
  }

  render() {
    const { FieldRenderer, FormRenderer } = this.props;
    let fields = [];
    const tooltip = (
      <Tooltip id="tooltip">{this.state.currentErr}</Tooltip>
    );
    _.forEach(this.props.fields, field => {
      fields = this.makeField(fields, field, tooltip, FieldRenderer);
    });

    // when a custom FormRenderer is supplied, use it
    if (FormRenderer) {
      return <FormRenderer>{fields}</FormRenderer>;
    }
    // render form
    else return <form>{fields}</form>;
  }
}

const FormBuilder = connect(
  (state) => {
  return {
    user: state.zlSite.user
  }
}
)(FormBuilderContainer)

module.exports = FormBuilder;
