//@flow

import _ from 'lodash/core';
import { connect } from 'react-redux';
import appSubscription from './../containers/app';
import pageSubscription from './../containers/page';
import { compose } from 'redux';

// Polyfill required to make require.ensure work on the server side
if(typeof require.ensure !== "function"){
  require.ensure = function(d, c) { c(require) };
}

var loadedComponents;

/* get component; either default or if theme provides */
export function getComponent(component, fallback = 'PgStandard') {
  let WrappedComponent;

  // check if React component
  if ('PgStandard' !== fallback && (component.startsWith("Bootstrap")
  || RegExp('^[a-z]').test(component))) {
    return require('./../components/ReactComponent');
  }

  // correct component name
  component = component.charAt(0).toUpperCase() + component.slice(1);
  component = component.replace(' ', '');

  _.forEach(loadedComponents, function(val, theme) {
    if (!_.isUndefined(val[component])) {
      WrappedComponent = val[component];
      return false; // exit function
    }
  });
  if (!WrappedComponent) {
    // use default (Webpack will load all in ./routes/ dir)
    try {
      if ('PgStandard' === fallback) {
        // if 'Page-pg' is fallback then we're looking for a page route
        WrappedComponent = require('./'+component);
      }
      else {
        // looking for a page component (module, node, widget, etc)
        WrappedComponent = require('./../components/'+component);
      }
    } catch (err) {
      if (fallback === component) {
        // this would never happen but...
        throw err;
      }
      // since we can't find component, use default page
      WrappedComponent = getComponent(fallback, fallback);
    }

  }
  return WrappedComponent;
}

/* loads store connnected container with wrapped App component */
function getThemeApp() {
  let mapStateToProps = (state) => {
    return {
      pages: state.pages,
      zlSite: state.zlSite,
      themeData: state.themeData,
      loadedComponents,
      getComponent
    };
  }
  // component name
  let component = 'App';

  let WrappedComponent = getComponent(component);

  const enhance = compose(
    connect(mapStateToProps),
    appSubscription
  )
  return enhance(WrappedComponent);
}

/* loads store connnected container with wrapped page component */
/* TODO: verify if we shouldn't save already loaded component and return that
instead of initiating a new copy. May be better optimized. */
function getThemePage(component, scsId) {
  let mapStateToProps = (state) => {
    let page = state.pages[scsId];
    let zsDefaultContent = state.zlSite.zsDefaultContent;
    let gtags = state.zlSite.gtags;
    const {
      isFetching,
      lastUpdated,
      contents
    } = state.pages[scsId] || {
      isFetching: true,
      contents: []
    }
    let contentList;
    if (Array.isArray(contents) && contents.length > 0) {
      contentList = [];
      contents.forEach(function(scId) {
        contentList.push(state.siteContents[scId]);
      });
    }
    return { page, contentList, zsDefaultContent, isFetching, lastUpdated,
      getComponent, gtags };
  }

  // get pg type component. Add 'pg-' so themes can separate page/route
  // component from component. If same name is used, will not have confusion
  // e.g., pg-Form.js vs Form.js
  let compName = `Pg${component.charAt(0).toUpperCase() + component.slice(1)}`;
  let WrappedComponent = getComponent(compName);

  const enhance = compose(
    connect(mapStateToProps),
    pageSubscription
  )
  return enhance(WrappedComponent, scsId);
}

function setupRoute(zl:any, req, res) {
  // keep copy of all components provided by theme
  if (process.env.APP_ENV !== 'browser' && 'development' === req.app.get('env')) {
    // server-side development uses babel-register and not bundle so zlSite is not an alias
    loadedComponents = require(`${process.cwd()}/${res.locals.zlSite.contentPath}/compDict.js`)();
  }
  else loadedComponents = require('zlSite/compDict.js')();

  // build child routes
  var childRoutes = [], indexRoute = {}, uniPages={};
  _.forEach(zl.pages, function(page) {
    // universal Id Url
    if (page.scsOptions && page.scsOptions.UniId) {
      uniPages[page.scsOptions.UniId] = '/';
      if (zl.zlSite.baseUrlPath) {
        uniPages[page.scsOptions.UniId] += zl.zlSite.baseUrlPath + '/';
      }
      uniPages[page.scsOptions.UniId] += page.path;
    }

    // determine if index
    if (page.scsId === zl.zlSite.homePage) {
      indexRoute = { component: getThemePage(page.scsBase, page.scsId) }
    }
    else {
      childRoutes.push({
        path: page.path,
        component: getThemePage(page.scsBase, page.scsId)
      });
    }

    // form/catalog that hasDataOf, create also edit/list component
    if (page.scsOptions && page.scsOptions.hasDataOf) {
      childRoutes.push({
        path: page.path+'/:id',
        component: getThemePage(page.scsBase, page.scsId)
      });
    }
  });

  if (process.env.APP_ENV === 'browser') {
    // make available in global scope
    window.uniPages = uniPages;
  }

  if (_.isEmpty(indexRoute) && childRoutes[0]) {
    // if no index was found, use first
    indexRoute = { component: childRoutes[0].component }
  }

  var rootRoute = {
    path: '/'+zl.zlSite.baseUrlPath,
    childRoutes: [
      {
        component: getThemeApp(),
        indexRoute: indexRoute,
        childRoutes: childRoutes
      }
    ]
  };
  return rootRoute;
};
export default setupRoute;
